# DO NOT EDIT THIS FILE!
# This file was generated by the R script
#     generate_fisher_exact_results_from_r.R
# The script was run with R version 3.6.2 (2019-12-12) at 2020-11-09 06:16:09


from collections import namedtuple
import numpy as np


Inf = np.inf

Parameters = namedtuple('Parameters',
                        ['table', 'confidence_level', 'alternative'])
RResults = namedtuple('RResults',
                      ['pvalue', 'conditional_odds_ratio',
                       'conditional_odds_ratio_ci'])
data = [
    (Parameters(table=[[100, 2], [1000, 5]],
                confidence_level=0.95,
                alternative='two.sided'),
     RResults(pvalue=0.1300759363430016,
              conditional_odds_ratio=0.25055839934223,
              conditional_odds_ratio_ci=(0.04035202926536294,
                                         2.662846672960251))),
    (Parameters(table=[[2, 7], [8, 2]],
                confidence_level=0.95,
                alternative='two.sided'),
     RResults(pvalue=0.02301413756522116,
              conditional_odds_ratio=0.0858623513573622,
              conditional_odds_ratio_ci=(0.004668988338943325,
                                         0.895792956493601))),
    (Parameters(table=[[5, 1], [10, 10]],
                confidence_level=0.95,
                alternative='two.sided'),
     RResults(pvalue=0.1973244147157191,
              conditional_odds_ratio=4.725646047336587,
              conditional_odds_ratio_ci=(0.4153910882532168,
                                         259.2593661129417))),
    (Parameters(table=[[5, 15], [20, 20]],
                confidence_level=0.95,
                alternative='two.sided'),
     RResults(pvalue=0.09580440012477633,
              conditional_odds_ratio=0.3394396617440851,
              conditional_odds_ratio_ci=(0.08056337526385809,
                                         1.22704788545557))),
    (Parameters(table=[[5, 16], [16, 25]],
                confidence_level=0.95,
                alternative='two.sided'),
     RResults(pvalue=0.2697004098849359,
              conditional_odds_ratio=0.4937791394540491,
              conditional_odds_ratio_ci=(0.1176691231650079,
                                         1.787463657995973))),
    (Parameters(table=[[10, 5], [10, 1]],
                confidence_level=0.95,
                alternative='two.sided'),
     RResults(pvalue=0.1973244147157192,
              conditional_odds_ratio=0.2116112781158479,
              conditional_odds_ratio_ci=(0.003857141267422399,
                                         2.407369893767229))),
    (Parameters(table=[[10, 5], [10, 0]],
                confidence_level=0.95,
                alternative='two.sided'),
     RResults(pvalue=0.06126482213438735,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         1.451643573543705))),
    (Parameters(table=[[5, 0], [1, 4]],
                confidence_level=0.95,
                alternative='two.sided'),
     RResults(pvalue=0.04761904761904762,
              conditional_odds_ratio=Inf,
              conditional_odds_ratio_ci=(1.024822256141754,
                                         Inf))),
    (Parameters(table=[[0, 5], [1, 4]],
                confidence_level=0.95,
                alternative='two.sided'),
     RResults(pvalue=1,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         39.00054996869288))),
    (Parameters(table=[[5, 1], [0, 4]],
                confidence_level=0.95,
                alternative='two.sided'),
     RResults(pvalue=0.04761904761904761,
              conditional_odds_ratio=Inf,
              conditional_odds_ratio_ci=(1.024822256141754,
                                         Inf))),
    (Parameters(table=[[0, 1], [3, 2]],
                confidence_level=0.95,
                alternative='two.sided'),
     RResults(pvalue=1,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         39.00054996869287))),
    (Parameters(table=[[200, 7], [8, 300]],
                confidence_level=0.95,
                alternative='two.sided'),
     RResults(pvalue=2.005657880389071e-122,
              conditional_odds_ratio=977.7866978606228,
              conditional_odds_ratio_ci=(349.2595113327733,
                                         3630.382605689872))),
    (Parameters(table=[[28, 21], [6, 1957]],
                confidence_level=0.95,
                alternative='two.sided'),
     RResults(pvalue=5.728437460831947e-44,
              conditional_odds_ratio=425.2403028434684,
              conditional_odds_ratio_ci=(152.4166024390096,
                                         1425.700792178893))),
    (Parameters(table=[[190, 800], [200, 900]],
                confidence_level=0.95,
                alternative='two.sided'),
     RResults(pvalue=0.574111858126088,
              conditional_odds_ratio=1.068697577856801,
              conditional_odds_ratio_ci=(0.8520462587912048,
                                         1.340148950273938))),
    (Parameters(table=[[100, 2], [1000, 5]],
                confidence_level=0.99,
                alternative='two.sided'),
     RResults(pvalue=0.1300759363430016,
              conditional_odds_ratio=0.25055839934223,
              conditional_odds_ratio_ci=(0.02502345007115455,
                                         6.304424772117853))),
    (Parameters(table=[[2, 7], [8, 2]],
                confidence_level=0.99,
                alternative='two.sided'),
     RResults(pvalue=0.02301413756522116,
              conditional_odds_ratio=0.0858623513573622,
              conditional_odds_ratio_ci=(0.001923034001462487,
                                         1.53670836950172))),
    (Parameters(table=[[5, 1], [10, 10]],
                confidence_level=0.99,
                alternative='two.sided'),
     RResults(pvalue=0.1973244147157191,
              conditional_odds_ratio=4.725646047336587,
              conditional_odds_ratio_ci=(0.2397970951413721,
                                         1291.342011095509))),
    (Parameters(table=[[5, 15], [20, 20]],
                confidence_level=0.99,
                alternative='two.sided'),
     RResults(pvalue=0.09580440012477633,
              conditional_odds_ratio=0.3394396617440851,
              conditional_odds_ratio_ci=(0.05127576113762925,
                                         1.717176678806983))),
    (Parameters(table=[[5, 16], [16, 25]],
                confidence_level=0.99,
                alternative='two.sided'),
     RResults(pvalue=0.2697004098849359,
              conditional_odds_ratio=0.4937791394540491,
              conditional_odds_ratio_ci=(0.07498546954483619,
                                         2.506969905199901))),
    (Parameters(table=[[10, 5], [10, 1]],
                confidence_level=0.99,
                alternative='two.sided'),
     RResults(pvalue=0.1973244147157192,
              conditional_odds_ratio=0.2116112781158479,
              conditional_odds_ratio_ci=(0.0007743881879531337,
                                         4.170192301163831))),
    (Parameters(table=[[10, 5], [10, 0]],
                confidence_level=0.99,
                alternative='two.sided'),
     RResults(pvalue=0.06126482213438735,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         2.642491011905582))),
    (Parameters(table=[[5, 0], [1, 4]],
                confidence_level=0.99,
                alternative='two.sided'),
     RResults(pvalue=0.04761904761904762,
              conditional_odds_ratio=Inf,
              conditional_odds_ratio_ci=(0.496935393325443,
                                         Inf))),
    (Parameters(table=[[0, 5], [1, 4]],
                confidence_level=0.99,
                alternative='two.sided'),
     RResults(pvalue=1,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         198.019801980198))),
    (Parameters(table=[[5, 1], [0, 4]],
                confidence_level=0.99,
                alternative='two.sided'),
     RResults(pvalue=0.04761904761904761,
              conditional_odds_ratio=Inf,
              conditional_odds_ratio_ci=(0.496935393325443,
                                         Inf))),
    (Parameters(table=[[0, 1], [3, 2]],
                confidence_level=0.99,
                alternative='two.sided'),
     RResults(pvalue=1,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         198.019801980198))),
    (Parameters(table=[[200, 7], [8, 300]],
                confidence_level=0.99,
                alternative='two.sided'),
     RResults(pvalue=2.005657880389071e-122,
              conditional_odds_ratio=977.7866978606228,
              conditional_odds_ratio_ci=(270.0334165523604,
                                         5461.333333326708))),
    (Parameters(table=[[28, 21], [6, 1957]],
                confidence_level=0.99,
                alternative='two.sided'),
     RResults(pvalue=5.728437460831947e-44,
              conditional_odds_ratio=425.2403028434684,
              conditional_odds_ratio_ci=(116.7944750275836,
                                         1931.995993191814))),
    (Parameters(table=[[190, 800], [200, 900]],
                confidence_level=0.99,
                alternative='two.sided'),
     RResults(pvalue=0.574111858126088,
              conditional_odds_ratio=1.068697577856801,
              conditional_odds_ratio_ci=(0.7949398282935892,
                                         1.436229679394333))),
    (Parameters(table=[[100, 2], [1000, 5]],
                confidence_level=0.95,
                alternative='less'),
     RResults(pvalue=0.1300759363430016,
              conditional_odds_ratio=0.25055839934223,
              conditional_odds_ratio_ci=(0,
                                         1.797867027270803))),
    (Parameters(table=[[2, 7], [8, 2]],
                confidence_level=0.95,
                alternative='less'),
     RResults(pvalue=0.0185217259520665,
              conditional_odds_ratio=0.0858623513573622,
              conditional_odds_ratio_ci=(0,
                                         0.6785254803404526))),
    (Parameters(table=[[5, 1], [10, 10]],
                confidence_level=0.95,
                alternative='less'),
     RResults(pvalue=0.9782608695652173,
              conditional_odds_ratio=4.725646047336587,
              conditional_odds_ratio_ci=(0,
                                         127.8497388102893))),
    (Parameters(table=[[5, 15], [20, 20]],
                confidence_level=0.95,
                alternative='less'),
     RResults(pvalue=0.05625775074399956,
              conditional_odds_ratio=0.3394396617440851,
              conditional_odds_ratio_ci=(0,
                                         1.032332939718425))),
    (Parameters(table=[[5, 16], [16, 25]],
                confidence_level=0.95,
                alternative='less'),
     RResults(pvalue=0.1808979350599346,
              conditional_odds_ratio=0.4937791394540491,
              conditional_odds_ratio_ci=(0,
                                         1.502407513296985))),
    (Parameters(table=[[10, 5], [10, 1]],
                confidence_level=0.95,
                alternative='less'),
     RResults(pvalue=0.1652173913043479,
              conditional_odds_ratio=0.2116112781158479,
              conditional_odds_ratio_ci=(0,
                                         1.820421051562392))),
    (Parameters(table=[[10, 5], [10, 0]],
                confidence_level=0.95,
                alternative='less'),
     RResults(pvalue=0.0565217391304348,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         1.06224603077045))),
    (Parameters(table=[[5, 0], [1, 4]],
                confidence_level=0.95,
                alternative='less'),
     RResults(pvalue=1,
              conditional_odds_ratio=Inf,
              conditional_odds_ratio_ci=(0,
                                         Inf))),
    (Parameters(table=[[0, 5], [1, 4]],
                confidence_level=0.95,
                alternative='less'),
     RResults(pvalue=0.5,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         19.00192394479939))),
    (Parameters(table=[[5, 1], [0, 4]],
                confidence_level=0.95,
                alternative='less'),
     RResults(pvalue=1,
              conditional_odds_ratio=Inf,
              conditional_odds_ratio_ci=(0,
                                         Inf))),
    (Parameters(table=[[0, 1], [3, 2]],
                confidence_level=0.95,
                alternative='less'),
     RResults(pvalue=0.4999999999999999,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         19.00192394479939))),
    (Parameters(table=[[200, 7], [8, 300]],
                confidence_level=0.95,
                alternative='less'),
     RResults(pvalue=1,
              conditional_odds_ratio=977.7866978606228,
              conditional_odds_ratio_ci=(0,
                                         3045.460216525746))),
    (Parameters(table=[[28, 21], [6, 1957]],
                confidence_level=0.95,
                alternative='less'),
     RResults(pvalue=1,
              conditional_odds_ratio=425.2403028434684,
              conditional_odds_ratio_ci=(0,
                                         1186.440170942579))),
    (Parameters(table=[[190, 800], [200, 900]],
                confidence_level=0.95,
                alternative='less'),
     RResults(pvalue=0.7416227010368963,
              conditional_odds_ratio=1.068697577856801,
              conditional_odds_ratio_ci=(0,
                                         1.293551891610822))),
    (Parameters(table=[[100, 2], [1000, 5]],
                confidence_level=0.99,
                alternative='less'),
     RResults(pvalue=0.1300759363430016,
              conditional_odds_ratio=0.25055839934223,
              conditional_odds_ratio_ci=(0,
                                         4.375946050832565))),
    (Parameters(table=[[2, 7], [8, 2]],
                confidence_level=0.99,
                alternative='less'),
     RResults(pvalue=0.0185217259520665,
              conditional_odds_ratio=0.0858623513573622,
              conditional_odds_ratio_ci=(0,
                                         1.235282118191202))),
    (Parameters(table=[[5, 1], [10, 10]],
                confidence_level=0.99,
                alternative='less'),
     RResults(pvalue=0.9782608695652173,
              conditional_odds_ratio=4.725646047336587,
              conditional_odds_ratio_ci=(0,
                                         657.2063583945989))),
    (Parameters(table=[[5, 15], [20, 20]],
                confidence_level=0.99,
                alternative='less'),
     RResults(pvalue=0.05625775074399956,
              conditional_odds_ratio=0.3394396617440851,
              conditional_odds_ratio_ci=(0,
                                         1.498867660683128))),
    (Parameters(table=[[5, 16], [16, 25]],
                confidence_level=0.99,
                alternative='less'),
     RResults(pvalue=0.1808979350599346,
              conditional_odds_ratio=0.4937791394540491,
              conditional_odds_ratio_ci=(0,
                                         2.186159386716762))),
    (Parameters(table=[[10, 5], [10, 1]],
                confidence_level=0.99,
                alternative='less'),
     RResults(pvalue=0.1652173913043479,
              conditional_odds_ratio=0.2116112781158479,
              conditional_odds_ratio_ci=(0,
                                         3.335351451901569))),
    (Parameters(table=[[10, 5], [10, 0]],
                confidence_level=0.99,
                alternative='less'),
     RResults(pvalue=0.0565217391304348,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         2.075407697450433))),
    (Parameters(table=[[5, 0], [1, 4]],
                confidence_level=0.99,
                alternative='less'),
     RResults(pvalue=1,
              conditional_odds_ratio=Inf,
              conditional_odds_ratio_ci=(0,
                                         Inf))),
    (Parameters(table=[[0, 5], [1, 4]],
                confidence_level=0.99,
                alternative='less'),
     RResults(pvalue=0.5,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         99.00009507969122))),
    (Parameters(table=[[5, 1], [0, 4]],
                confidence_level=0.99,
                alternative='less'),
     RResults(pvalue=1,
              conditional_odds_ratio=Inf,
              conditional_odds_ratio_ci=(0,
                                         Inf))),
    (Parameters(table=[[0, 1], [3, 2]],
                confidence_level=0.99,
                alternative='less'),
     RResults(pvalue=0.4999999999999999,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         99.00009507969123))),
    (Parameters(table=[[200, 7], [8, 300]],
                confidence_level=0.99,
                alternative='less'),
     RResults(pvalue=1,
              conditional_odds_ratio=977.7866978606228,
              conditional_odds_ratio_ci=(0,
                                         4503.078257659934))),
    (Parameters(table=[[28, 21], [6, 1957]],
                confidence_level=0.99,
                alternative='less'),
     RResults(pvalue=1,
              conditional_odds_ratio=425.2403028434684,
              conditional_odds_ratio_ci=(0,
                                         1811.766127544222))),
    (Parameters(table=[[190, 800], [200, 900]],
                confidence_level=0.99,
                alternative='less'),
     RResults(pvalue=0.7416227010368963,
              conditional_odds_ratio=1.068697577856801,
              conditional_odds_ratio_ci=(0,
                                         1.396522811516685))),
    (Parameters(table=[[100, 2], [1000, 5]],
                confidence_level=0.95,
                alternative='greater'),
     RResults(pvalue=0.979790445314723,
              conditional_odds_ratio=0.25055839934223,
              conditional_odds_ratio_ci=(0.05119649909830196,
                                         Inf))),
    (Parameters(table=[[2, 7], [8, 2]],
                confidence_level=0.95,
                alternative='greater'),
     RResults(pvalue=0.9990149169715733,
              conditional_odds_ratio=0.0858623513573622,
              conditional_odds_ratio_ci=(0.007163749169069961,
                                         Inf))),
    (Parameters(table=[[5, 1], [10, 10]],
                confidence_level=0.95,
                alternative='greater'),
     RResults(pvalue=0.1652173913043478,
              conditional_odds_ratio=4.725646047336587,
              conditional_odds_ratio_ci=(0.5493234651081089,
                                         Inf))),
    (Parameters(table=[[5, 15], [20, 20]],
                confidence_level=0.95,
                alternative='greater'),
     RResults(pvalue=0.9849086665340765,
              conditional_odds_ratio=0.3394396617440851,
              conditional_odds_ratio_ci=(0.1003538933958604,
                                         Inf))),
    (Parameters(table=[[5, 16], [16, 25]],
                confidence_level=0.95,
                alternative='greater'),
     RResults(pvalue=0.9330176609214881,
              conditional_odds_ratio=0.4937791394540491,
              conditional_odds_ratio_ci=(0.146507416280863,
                                         Inf))),
    (Parameters(table=[[10, 5], [10, 1]],
                confidence_level=0.95,
                alternative='greater'),
     RResults(pvalue=0.9782608695652174,
              conditional_odds_ratio=0.2116112781158479,
              conditional_odds_ratio_ci=(0.007821681994077808,
                                         Inf))),
    (Parameters(table=[[10, 5], [10, 0]],
                confidence_level=0.95,
                alternative='greater'),
     RResults(pvalue=1,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         Inf))),
    (Parameters(table=[[5, 0], [1, 4]],
                confidence_level=0.95,
                alternative='greater'),
     RResults(pvalue=0.02380952380952382,
              conditional_odds_ratio=Inf,
              conditional_odds_ratio_ci=(1.487678929918272,
                                         Inf))),
    (Parameters(table=[[0, 5], [1, 4]],
                confidence_level=0.95,
                alternative='greater'),
     RResults(pvalue=1,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         Inf))),
    (Parameters(table=[[5, 1], [0, 4]],
                confidence_level=0.95,
                alternative='greater'),
     RResults(pvalue=0.0238095238095238,
              conditional_odds_ratio=Inf,
              conditional_odds_ratio_ci=(1.487678929918272,
                                         Inf))),
    (Parameters(table=[[0, 1], [3, 2]],
                confidence_level=0.95,
                alternative='greater'),
     RResults(pvalue=1,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         Inf))),
    (Parameters(table=[[200, 7], [8, 300]],
                confidence_level=0.95,
                alternative='greater'),
     RResults(pvalue=2.005657880388915e-122,
              conditional_odds_ratio=977.7866978606228,
              conditional_odds_ratio_ci=(397.784359748113,
                                         Inf))),
    (Parameters(table=[[28, 21], [6, 1957]],
                confidence_level=0.95,
                alternative='greater'),
     RResults(pvalue=5.728437460831983e-44,
              conditional_odds_ratio=425.2403028434684,
              conditional_odds_ratio_ci=(174.7148056880929,
                                         Inf))),
    (Parameters(table=[[190, 800], [200, 900]],
                confidence_level=0.95,
                alternative='greater'),
     RResults(pvalue=0.2959825901308897,
              conditional_odds_ratio=1.068697577856801,
              conditional_odds_ratio_ci=(0.8828406663967776,
                                         Inf))),
    (Parameters(table=[[100, 2], [1000, 5]],
                confidence_level=0.99,
                alternative='greater'),
     RResults(pvalue=0.979790445314723,
              conditional_odds_ratio=0.25055839934223,
              conditional_odds_ratio_ci=(0.03045407081240429,
                                         Inf))),
    (Parameters(table=[[2, 7], [8, 2]],
                confidence_level=0.99,
                alternative='greater'),
     RResults(pvalue=0.9990149169715733,
              conditional_odds_ratio=0.0858623513573622,
              conditional_odds_ratio_ci=(0.002768053063547901,
                                         Inf))),
    (Parameters(table=[[5, 1], [10, 10]],
                confidence_level=0.99,
                alternative='greater'),
     RResults(pvalue=0.1652173913043478,
              conditional_odds_ratio=4.725646047336587,
              conditional_odds_ratio_ci=(0.2998184792279909,
                                         Inf))),
    (Parameters(table=[[5, 15], [20, 20]],
                confidence_level=0.99,
                alternative='greater'),
     RResults(pvalue=0.9849086665340765,
              conditional_odds_ratio=0.3394396617440851,
              conditional_odds_ratio_ci=(0.06180414342643172,
                                         Inf))),
    (Parameters(table=[[5, 16], [16, 25]],
                confidence_level=0.99,
                alternative='greater'),
     RResults(pvalue=0.9330176609214881,
              conditional_odds_ratio=0.4937791394540491,
              conditional_odds_ratio_ci=(0.09037094010066403,
                                         Inf))),
    (Parameters(table=[[10, 5], [10, 1]],
                confidence_level=0.99,
                alternative='greater'),
     RResults(pvalue=0.9782608695652174,
              conditional_odds_ratio=0.2116112781158479,
              conditional_odds_ratio_ci=(0.001521592095430679,
                                         Inf))),
    (Parameters(table=[[10, 5], [10, 0]],
                confidence_level=0.99,
                alternative='greater'),
     RResults(pvalue=1,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         Inf))),
    (Parameters(table=[[5, 0], [1, 4]],
                confidence_level=0.99,
                alternative='greater'),
     RResults(pvalue=0.02380952380952382,
              conditional_odds_ratio=Inf,
              conditional_odds_ratio_ci=(0.6661157890359722,
                                         Inf))),
    (Parameters(table=[[0, 5], [1, 4]],
                confidence_level=0.99,
                alternative='greater'),
     RResults(pvalue=1,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         Inf))),
    (Parameters(table=[[5, 1], [0, 4]],
                confidence_level=0.99,
                alternative='greater'),
     RResults(pvalue=0.0238095238095238,
              conditional_odds_ratio=Inf,
              conditional_odds_ratio_ci=(0.6661157890359725,
                                         Inf))),
    (Parameters(table=[[0, 1], [3, 2]],
                confidence_level=0.99,
                alternative='greater'),
     RResults(pvalue=1,
              conditional_odds_ratio=0,
              conditional_odds_ratio_ci=(0,
                                         Inf))),
    (Parameters(table=[[200, 7], [8, 300]],
                confidence_level=0.99,
                alternative='greater'),
     RResults(pvalue=2.005657880388915e-122,
              conditional_odds_ratio=977.7866978606228,
              conditional_odds_ratio_ci=(297.9619252357688,
                                         Inf))),
    (Parameters(table=[[28, 21], [6, 1957]],
                confidence_level=0.99,
                alternative='greater'),
     RResults(pvalue=5.728437460831983e-44,
              conditional_odds_ratio=425.2403028434684,
              conditional_odds_ratio_ci=(130.3213490295859,
                                         Inf))),
    (Parameters(table=[[190, 800], [200, 900]],
                confidence_level=0.99,
                alternative='greater'),
     RResults(pvalue=0.2959825901308897,
              conditional_odds_ratio=1.068697577856801,
              conditional_odds_ratio_ci=(0.8176272148267533,
                                         Inf))),
]



"""Test that all ufuncs have float32-preserving signatures.

This was once guaranteed through the code generation script for
generating ufuncs, `scipy/special/_generate_pyx.py`. Starting with
gh-20260, SciPy developers have begun moving to generate ufuncs
through direct use of the NumPy C API (through C++). Existence of
float32 preserving signatures must now be tested since it is no
longer guaranteed.
"""

import numpy as np
import pytest
import scipy.special._ufuncs
import scipy.special._gufuncs

_ufuncs = []
for funcname in dir(scipy.special._ufuncs):
    _ufuncs.append(getattr(scipy.special._ufuncs, funcname))
for funcname in dir(scipy.special._gufuncs):
    _ufuncs.append(getattr(scipy.special._gufuncs, funcname))

# Not all module members are actually ufuncs
_ufuncs = [func for func in _ufuncs if isinstance(func, np.ufunc)]

@pytest.mark.parametrize("ufunc", _ufuncs)
def test_ufunc_signatures(ufunc):

    # From _generate_pyx.py
    # "Don't add float32 versions of ufuncs with integer arguments, as this
    # can lead to incorrect dtype selection if the integer arguments are
    # arrays, but float arguments are scalars.
    # For instance sph_harm(0,[0],0,0).dtype == complex64
    # This may be a NumPy bug, but we need to work around it.
    # cf. gh-4895, https://github.com/numpy/numpy/issues/5895"
    types = set(sig for sig in ufunc.types
                if not ("l" in sig or "i" in sig or "q" in sig or "p" in sig))

    # Generate the full expanded set of signatures which should exist. There
    # should be matching float and double versions of any existing signature.
    expanded_types = set()
    for sig in types:
        expanded_types.update(
            [sig.replace("d", "f").replace("D", "F"),
             sig.replace("f", "d").replace("F", "D")]
        )
    assert types == expanded_types



import collections
from abc import ABC, abstractmethod

import numpy as np

from scipy._lib._util import MapWrapper


class VertexBase(ABC):
    """
    Base class for a vertex.
    """
    def __init__(self, x, nn=None, index=None):
        """
        Initiation of a vertex object.

        Parameters
        ----------
        x : tuple or vector
            The geometric location (domain).
        nn : list, optional
            Nearest neighbour list.
        index : int, optional
            Index of vertex.
        """
        self.x = x
        self.hash = hash(self.x)  # Save precomputed hash

        if nn is not None:
            self.nn = set(nn)  # can use .indexupdate to add a new list
        else:
            self.nn = set()

        self.index = index

    def __hash__(self):
        return self.hash

    def __getattr__(self, item):
        if item not in ['x_a']:
            raise AttributeError(f"{type(self)} object has no attribute "
                                 f"'{item}'")
        if item == 'x_a':
            self.x_a = np.array(self.x)
            return self.x_a

    @abstractmethod
    def connect(self, v):
        raise NotImplementedError("This method is only implemented with an "
                                  "associated child of the base class.")

    @abstractmethod
    def disconnect(self, v):
        raise NotImplementedError("This method is only implemented with an "
                                  "associated child of the base class.")

    def star(self):
        """Returns the star domain ``st(v)`` of the vertex.

        Parameters
        ----------
        v :
            The vertex ``v`` in ``st(v)``

        Returns
        -------
        st : set
            A set containing all the vertices in ``st(v)``
        """
        self.st = self.nn
        self.st.add(self)
        return self.st


class VertexScalarField(VertexBase):
    """
    Add homology properties of a scalar field f: R^n --> R associated with
    the geometry built from the VertexBase class
    """

    def __init__(self, x, field=None, nn=None, index=None, field_args=(),
                 g_cons=None, g_cons_args=()):
        """
        Parameters
        ----------
        x : tuple,
            vector of vertex coordinates
        field : callable, optional
            a scalar field f: R^n --> R associated with the geometry
        nn : list, optional
            list of nearest neighbours
        index : int, optional
            index of the vertex
        field_args : tuple, optional
            additional arguments to be passed to field
        g_cons : callable, optional
            constraints on the vertex
        g_cons_args : tuple, optional
            additional arguments to be passed to g_cons

        """
        super().__init__(x, nn=nn, index=index)

        # Note Vertex is only initiated once for all x so only
        # evaluated once
        # self.feasible = None

        # self.f is externally defined by the cache to allow parallel
        # processing
        # None type that will break arithmetic operations unless defined
        # self.f = None

        self.check_min = True
        self.check_max = True

    def connect(self, v):
        """Connects self to another vertex object v.

        Parameters
        ----------
        v : VertexBase or VertexScalarField object
        """
        if v is not self and v not in self.nn:
            self.nn.add(v)
            v.nn.add(self)

            # Flags for checking homology properties:
            self.check_min = True
            self.check_max = True
            v.check_min = True
            v.check_max = True

    def disconnect(self, v):
        if v in self.nn:
            self.nn.remove(v)
            v.nn.remove(self)

            # Flags for checking homology properties:
            self.check_min = True
            self.check_max = True
            v.check_min = True
            v.check_max = True

    def minimiser(self):
        """Check whether this vertex is strictly less than all its
           neighbours"""
        if self.check_min:
            self._min = all(self.f < v.f for v in self.nn)
            self.check_min = False

        return self._min

    def maximiser(self):
        """
        Check whether this vertex is strictly greater than all its
        neighbours.
        """
        if self.check_max:
            self._max = all(self.f > v.f for v in self.nn)
            self.check_max = False

        return self._max


class VertexVectorField(VertexBase):
    """
    Add homology properties of a scalar field f: R^n --> R^m associated with
    the geometry built from the VertexBase class.
    """

    def __init__(self, x, sfield=None, vfield=None, field_args=(),
                 vfield_args=(), g_cons=None,
                 g_cons_args=(), nn=None, index=None):
        super().__init__(x, nn=nn, index=index)

        raise NotImplementedError("This class is still a work in progress")


class VertexCacheBase:
    """Base class for a vertex cache for a simplicial complex."""
    def __init__(self):

        self.cache = collections.OrderedDict()
        self.nfev = 0  # Feasible points
        self.index = -1

    def __iter__(self):
        for v in self.cache:
            yield self.cache[v]
        return

    def size(self):
        """Returns the size of the vertex cache."""
        return self.index + 1

    def print_out(self):
        headlen = len(f"Vertex cache of size: {len(self.cache)}:")
        print('=' * headlen)
        print(f"Vertex cache of size: {len(self.cache)}:")
        print('=' * headlen)
        for v in self.cache:
            self.cache[v].print_out()


class VertexCube(VertexBase):
    """Vertex class to be used for a pure simplicial complex with no associated
    differential geometry (single level domain that exists in R^n)"""
    def __init__(self, x, nn=None, index=None):
        super().__init__(x, nn=nn, index=index)

    def connect(self, v):
        if v is not self and v not in self.nn:
            self.nn.add(v)
            v.nn.add(self)

    def disconnect(self, v):
        if v in self.nn:
            self.nn.remove(v)
            v.nn.remove(self)


class VertexCacheIndex(VertexCacheBase):
    def __init__(self):
        """
        Class for a vertex cache for a simplicial complex without an associated
        field. Useful only for building and visualising a domain complex.

        Parameters
        ----------
        """
        super().__init__()
        self.Vertex = VertexCube

    def __getitem__(self, x, nn=None):
        try:
            return self.cache[x]
        except KeyError:
            self.index += 1
            xval = self.Vertex(x, index=self.index)
            # logging.info("New generated vertex at x = {}".format(x))
            # NOTE: Surprisingly high performance increase if logging
            # is commented out
            self.cache[x] = xval
            return self.cache[x]


class VertexCacheField(VertexCacheBase):
    def __init__(self, field=None, field_args=(), g_cons=None, g_cons_args=(),
                 workers=1):
        """
        Class for a vertex cache for a simplicial complex with an associated
        field.

        Parameters
        ----------
        field : callable
            Scalar or vector field callable.
        field_args : tuple, optional
            Any additional fixed parameters needed to completely specify the
            field function
        g_cons : dict or sequence of dict, optional
            Constraints definition.
            Function(s) ``R**n`` in the form::
        g_cons_args : tuple, optional
            Any additional fixed parameters needed to completely specify the
            constraint functions
        workers : int  optional
            Uses `multiprocessing.Pool <multiprocessing>`) to compute the field
             functions in parallel.

        """
        super().__init__()
        self.index = -1
        self.Vertex = VertexScalarField
        self.field = field
        self.field_args = field_args
        self.wfield = FieldWrapper(field, field_args)  # if workers is not 1

        self.g_cons = g_cons
        self.g_cons_args = g_cons_args
        self.wgcons = ConstraintWrapper(g_cons, g_cons_args)
        self.gpool = set()  # A set of tuples to process for feasibility

        # Field processing objects
        self.fpool = set()  # A set of tuples to process for scalar function
        self.sfc_lock = False  # True if self.fpool is non-Empty

        self.workers = workers
        self._mapwrapper = MapWrapper(workers)

        if workers == 1:
            self.process_gpool = self.proc_gpool
            if g_cons is None:
                self.process_fpool = self.proc_fpool_nog
            else:
                self.process_fpool = self.proc_fpool_g
        else:
            self.process_gpool = self.pproc_gpool
            if g_cons is None:
                self.process_fpool = self.pproc_fpool_nog
            else:
                self.process_fpool = self.pproc_fpool_g

    def __getitem__(self, x, nn=None):
        try:
            return self.cache[x]
        except KeyError:
            self.index += 1
            xval = self.Vertex(x, field=self.field, nn=nn, index=self.index,
                               field_args=self.field_args,
                               g_cons=self.g_cons,
                               g_cons_args=self.g_cons_args)

            self.cache[x] = xval  # Define in cache
            self.gpool.add(xval)  # Add to pool for processing feasibility
            self.fpool.add(xval)  # Add to pool for processing field values
            return self.cache[x]

    def __getstate__(self):
        self_dict = self.__dict__.copy()
        del self_dict['pool']
        return self_dict

    def process_pools(self):
        if self.g_cons is not None:
            self.process_gpool()
        self.process_fpool()
        self.proc_minimisers()

    def feasibility_check(self, v):
        v.feasible = True
        for g, args in zip(self.g_cons, self.g_cons_args):
            # constraint may return more than 1 value.
            if np.any(g(v.x_a, *args) < 0.0):
                v.f = np.inf
                v.feasible = False
                break

    def compute_sfield(self, v):
        """Compute the scalar field values of a vertex object `v`.

        Parameters
        ----------
        v : VertexBase or VertexScalarField object
        """
        try:
            v.f = self.field(v.x_a, *self.field_args)
            self.nfev += 1
        except AttributeError:
            v.f = np.inf
            # logging.warning(f"Field function not found at x = {self.x_a}")
        if np.isnan(v.f):
            v.f = np.inf

    def proc_gpool(self):
        """Process all constraints."""
        if self.g_cons is not None:
            for v in self.gpool:
                self.feasibility_check(v)
        # Clean the pool
        self.gpool = set()

    def pproc_gpool(self):
        """Process all constraints in parallel."""
        gpool_l = []
        for v in self.gpool:
            gpool_l.append(v.x_a)

        G = self._mapwrapper(self.wgcons.gcons, gpool_l)
        for v, g in zip(self.gpool, G):
            v.feasible = g  # set vertex object attribute v.feasible = g (bool)

    def proc_fpool_g(self):
        """Process all field functions with constraints supplied."""
        for v in self.fpool:
            if v.feasible:
                self.compute_sfield(v)
        # Clean the pool
        self.fpool = set()

    def proc_fpool_nog(self):
        """Process all field functions with no constraints supplied."""
        for v in self.fpool:
            self.compute_sfield(v)
        # Clean the pool
        self.fpool = set()

    def pproc_fpool_g(self):
        """
        Process all field functions with constraints supplied in parallel.
        """
        self.wfield.func
        fpool_l = []
        for v in self.fpool:
            if v.feasible:
                fpool_l.append(v.x_a)
            else:
                v.f = np.inf
        F = self._mapwrapper(self.wfield.func, fpool_l)
        for va, f in zip(fpool_l, F):
            vt = tuple(va)
            self[vt].f = f  # set vertex object attribute v.f = f
            self.nfev += 1
        # Clean the pool
        self.fpool = set()

    def pproc_fpool_nog(self):
        """
        Process all field functions with no constraints supplied in parallel.
        """
        self.wfield.func
        fpool_l = []
        for v in self.fpool:
            fpool_l.append(v.x_a)
        F = self._mapwrapper(self.wfield.func, fpool_l)
        for va, f in zip(fpool_l, F):
            vt = tuple(va)
            self[vt].f = f  # set vertex object attribute v.f = f
            self.nfev += 1
        # Clean the pool
        self.fpool = set()

    def proc_minimisers(self):
        """Check for minimisers."""
        for v in self:
            v.minimiser()
            v.maximiser()


class ConstraintWrapper:
    """Object to wrap constraints to pass to `multiprocessing.Pool`."""
    def __init__(self, g_cons, g_cons_args):
        self.g_cons = g_cons
        self.g_cons_args = g_cons_args

    def gcons(self, v_x_a):
        vfeasible = True
        for g, args in zip(self.g_cons, self.g_cons_args):
            # constraint may return more than 1 value.
            if np.any(g(v_x_a, *args) < 0.0):
                vfeasible = False
                break
        return vfeasible


class FieldWrapper:
    """Object to wrap field to pass to `multiprocessing.Pool`."""
    def __init__(self, field, field_args):
        self.field = field
        self.field_args = field_args

    def func(self, v_x_a):
        try:
            v_f = self.field(v_x_a, *self.field_args)
        except Exception:
            v_f = np.inf
        if np.isnan(v_f):
            v_f = np.inf

        return v_f



import math

import pytest
import numpy as np
import array
from numpy.testing import assert_equal

from array_api_compat import (  # noqa: F401
    is_numpy_array, is_cupy_array, is_torch_array,
    is_dask_array, is_jax_array, is_pydata_sparse_array,
    is_ndonnx_array,
    is_numpy_namespace, is_cupy_namespace, is_torch_namespace,
    is_dask_namespace, is_jax_namespace, is_pydata_sparse_namespace,
    is_array_api_strict_namespace, is_ndonnx_namespace,
)

from array_api_compat import (
    device, is_array_api_obj, is_lazy_array, is_writeable_array, size, to_device
)
from array_api_compat.common._helpers import _DASK_DEVICE
from ._helpers import all_libraries, import_, wrapped_libraries, xfail


is_array_functions = {
    'numpy': 'is_numpy_array',
    'cupy': 'is_cupy_array',
    'torch': 'is_torch_array',
    'dask.array': 'is_dask_array',
    'jax.numpy': 'is_jax_array',
    'sparse': 'is_pydata_sparse_array',
    'ndonnx': 'is_ndonnx_array',
}

is_namespace_functions = {
    'numpy': 'is_numpy_namespace',
    'cupy': 'is_cupy_namespace',
    'torch': 'is_torch_namespace',
    'dask.array': 'is_dask_namespace',
    'jax.numpy': 'is_jax_namespace',
    'sparse': 'is_pydata_sparse_namespace',
    'array_api_strict': 'is_array_api_strict_namespace',
    'ndonnx': 'is_ndonnx_namespace',
}


@pytest.mark.parametrize('library', is_array_functions.keys())
@pytest.mark.parametrize('func', is_array_functions.values())
def test_is_xp_array(library, func):
    lib = import_(library)
    is_func = globals()[func]

    x = lib.asarray([1, 2, 3])

    assert is_func(x) == (func == is_array_functions[library])

    assert is_array_api_obj(x)


@pytest.mark.parametrize('library', is_namespace_functions.keys())
@pytest.mark.parametrize('func', is_namespace_functions.values())
def test_is_xp_namespace(library, func):
    lib = import_(library)
    is_func = globals()[func]

    assert is_func(lib) == (func == is_namespace_functions[library])


@pytest.mark.parametrize('library', all_libraries)
def test_xp_is_array_generics(library):
    """
    Test that scalar selection on a xp.ndarray always returns
    an object that matches with exactly one among the is_*_array
    function of the same library and is_numpy_array.
    """
    lib = import_(library)
    x = lib.asarray([1, 2, 3])
    x0 = x[0]

    matches = []
    for library2, func in is_array_functions.items():
        is_func = globals()[func]
        if is_func(x0):
            matches.append(library2)

    if library == "array_api_strict":
        # There is no is_array_api_strict_array() function
        assert matches == []
    else:
        assert matches in ([library], ["numpy"])


@pytest.mark.parametrize("library", all_libraries)
def test_is_writeable_array(library):
    lib = import_(library)
    x = lib.asarray([1, 2, 3])
    if is_writeable_array(x):
        x[1] = 4
    else:
        with pytest.raises((TypeError, ValueError)):
            x[1] = 4


def test_is_writeable_array_numpy():
    x = np.asarray([1, 2, 3])
    assert is_writeable_array(x)
    x.flags.writeable = False
    assert not is_writeable_array(x)


@pytest.mark.parametrize("library", all_libraries)
def test_size(library):
    xp = import_(library)
    x = xp.asarray([1, 2, 3])
    assert size(x) == 3


@pytest.mark.parametrize("library", all_libraries)
def test_size_none(library):
    if library == "sparse":
        pytest.skip("No arange(); no indexing by sparse arrays")

    xp = import_(library)
    x = xp.arange(10)
    x = x[x < 5]

    # dask.array now has shape=(nan, ) and size=nan
    # ndonnx now has shape=(None, ) and size=None
    # Eager libraries have shape=(5, ) and size=5
    assert size(x) in (None, 5)


@pytest.mark.parametrize("library", all_libraries)
def test_is_lazy_array(library):
    lib = import_(library)
    x = lib.asarray([1, 2, 3])
    assert isinstance(is_lazy_array(x), bool)


@pytest.mark.parametrize("shape", [(math.nan,), (1, math.nan), (None, ), (1, None)])
def test_is_lazy_array_nan_size(shape, monkeypatch):
    """Test is_lazy_array() on an unknown Array API compliant object
    with NaN (like Dask) or None (like ndonnx) in its shape
    """
    xp = import_("array_api_strict")
    x = xp.asarray(1)
    assert not is_lazy_array(x)
    monkeypatch.setattr(type(x), "shape", shape)
    assert is_lazy_array(x)


@pytest.mark.parametrize("exc", [TypeError, AssertionError])
def test_is_lazy_array_bool_raises(exc, monkeypatch):
    """Test is_lazy_array() on an unknown Array API compliant object
    where calling bool() raises:
    - TypeError: e.g. like jitted JAX. This is the proper exception which
      lazy arrays should raise as per the Array API specification
    - something else: e.g. like Dask, where bool() triggers compute()
      which can result in any kind of exception to be raised
    """
    xp = import_("array_api_strict")
    x = xp.asarray(1)
    assert not is_lazy_array(x)

    def __bool__(self):
        raise exc("Hello world")

    monkeypatch.setattr(type(x), "__bool__", __bool__)
    assert is_lazy_array(x)


@pytest.mark.parametrize(
    'func',
    list(is_array_functions.values()) 
    + ["is_array_api_obj", "is_lazy_array", "is_writeable_array"]
)
def test_is_array_any_object(func):
    """Test that is_*_array functions return False and don't raise on non-array objects
    """
    func = globals()[func]

    # These objects are missing attributes such as __name__
    assert not func(object())
    assert not func(None)
    assert not func(1)

    class C:
        pass

    assert not func(C())


@pytest.mark.parametrize("library", all_libraries)
def test_device_to_device(library, request):
    if library == "ndonnx":
        xfail(request, reason="Stub raises ValueError")
    if library == "sparse":
        xfail(request, reason="No __array_namespace_info__()")

    xp = import_(library, wrapper=True)
    devices = xp.__array_namespace_info__().devices()

    # Default device
    x = xp.asarray([1, 2, 3])
    dev = device(x)

    for dev in devices:
        if dev is None:  # JAX >=0.5.3
            continue
        if dev is _DASK_DEVICE:  # TODO this needs a better design
            continue
        y = to_device(x, dev)
        assert device(y) == dev


@pytest.mark.parametrize("library", wrapped_libraries)
def test_to_device_host(library):
    # different libraries have different semantics
    # for DtoH transfers; ensure that we support a portable
    # shim for common array libs
    # see: https://github.com/scipy/scipy/issues/18286#issuecomment-1527552919
    xp = import_(library, wrapper=True)

    expected = np.array([1, 2, 3])
    x = xp.asarray([1, 2, 3])
    x = to_device(x, "cpu")
    # torch will return a genuine Device object, but
    # the other libs will do something different with
    # a `device(x)` query; however, what's really important
    # here is that we can test portably after calling
    # to_device(x, "cpu") to return to host
    assert_equal(x, expected)


@pytest.mark.parametrize("target_library", is_array_functions.keys())
@pytest.mark.parametrize("source_library", is_array_functions.keys())
def test_asarray_cross_library(source_library, target_library, request):
    if source_library == "dask.array" and target_library == "torch":
        # TODO: remove xfail once
        # https://github.com/dask/dask/issues/8260 is resolved
        xfail(request, reason="Bug in dask raising error on conversion")

    elif (
        source_library == "ndonnx" 
        and target_library not in ("array_api_strict", "ndonnx", "numpy")
    ):
        xfail(request, reason="The truth value of lazy Array Array(dtype=Boolean) is unknown")
    elif source_library == "ndonnx" and target_library == "numpy":
        xfail(request, reason="produces numpy array of ndonnx scalar arrays")
    elif target_library == "ndonnx" and source_library in ("torch", "dask.array", "jax.numpy"):
        xfail(request, reason="unable to infer dtype")

    elif source_library == "jax.numpy" and target_library == "torch":
        xfail(request, reason="casts int to float")
    elif source_library == "cupy" and target_library != "cupy":
        # cupy explicitly disallows implicit conversions to CPU
        pytest.skip(reason="cupy does not support implicit conversion to CPU")
    elif source_library == "sparse" and target_library != "sparse":
        pytest.skip(reason="`sparse` does not allow implicit densification")

    src_lib = import_(source_library, wrapper=True)
    tgt_lib = import_(target_library, wrapper=True)
    is_tgt_type = globals()[is_array_functions[target_library]]

    a = src_lib.asarray([1, 2, 3], dtype=src_lib.int32)
    b = tgt_lib.asarray(a)

    assert is_tgt_type(b), f"Expected {b} to be a {tgt_lib.ndarray}, but was {type(b)}"
    assert b.dtype == tgt_lib.int32


@pytest.mark.parametrize("library", wrapped_libraries)
def test_asarray_copy(library):
    # Note, we have this test here because the test suite currently doesn't
    # test the copy flag to asarray() very rigorously. Once
    # https://github.com/data-apis/array-api-tests/issues/241 is fixed we
    # should be able to delete this.
    xp = import_(library, wrapper=True)
    asarray = xp.asarray
    is_lib_func = globals()[is_array_functions[library]]

    a = asarray([1])
    b = asarray(a, copy=True)
    assert is_lib_func(b)
    a[0] = 0
    assert b[0] == 1
    assert a[0] == 0

    a = asarray([1])

    # Test copy=False within the same namespace
    b = asarray(a, copy=False)
    assert is_lib_func(b)
    a[0] = 0
    assert b[0] == 0
    with pytest.raises(ValueError):
        asarray(a, copy=False, dtype=xp.float64)

    # copy=None defaults to False when possible
    a = asarray([1])
    b = asarray(a, copy=None)
    assert is_lib_func(b)
    a[0] = 0
    assert b[0] == 0

    # copy=None defaults to True when impossible
    a = asarray([1.0], dtype=xp.float32)
    assert a.dtype == xp.float32
    b = asarray(a, dtype=xp.float64, copy=None)
    assert is_lib_func(b)
    assert b.dtype == xp.float64
    a[0] = 0.0
    assert b[0] == 1.0

    # copy=None defaults to False when possible
    a = asarray([1.0], dtype=xp.float64)
    assert a.dtype == xp.float64
    b = asarray(a, dtype=xp.float64, copy=None)
    assert is_lib_func(b)
    assert b.dtype == xp.float64
    a[0] = 0.0
    assert b[0] == 0.0

    # Python built-in types
    for obj in [True, 0, 0.0, 0j, [0], [[0]]]:
        asarray(obj, copy=True)  # No error
        asarray(obj, copy=None)  # No error

        with pytest.raises(ValueError):
            asarray(obj, copy=False)

    # Use the standard library array to test the buffer protocol
    a = array.array("f", [1.0])
    b = asarray(a, copy=True)
    assert is_lib_func(b)
    a[0] = 0.0
    assert b[0] == 1.0

    a = array.array("f", [1.0])
    if library in ("cupy", "dask.array"):
        with pytest.raises(ValueError):
            asarray(a, copy=False)
    else:
        b = asarray(a, copy=False)
        assert is_lib_func(b)
        a[0] = 0.0
        assert b[0] == 0.0

    a = array.array("f", [1.0])
    b = asarray(a, copy=None)
    assert is_lib_func(b)
    a[0] = 0.0
    if library in ("cupy", "dask.array"):
        # A copy is required for libraries where the default device is not CPU
        # dask changed behaviour of copy=None in 2024.12 to copy;
        # this wrapper ensures the same behaviour in older versions too.
        # https://github.com/dask/dask/pull/11524/
        assert b[0] == 1.0
    else:
        # copy=None defaults to False when possible
        assert b[0] == 0.0


@pytest.mark.parametrize("library", ["numpy", "cupy", "torch"])
def test_clip_out(library):
    """Test non-standard out= parameter for clip()

    (see "Avoid Restricting Behavior that is Outside the Scope of the Standard"
    in https://data-apis.org/array-api-compat/dev/special-considerations.html)
    """
    xp = import_(library, wrapper=True)
    x = xp.asarray([10, 20, 30])
    out = xp.zeros_like(x)
    xp.clip(x, 15, 25, out=out)
    expect = xp.asarray([15, 20, 25])
    assert xp.all(out == expect)



cdef public int foo() noexcept:
    return 1

cdef public int foo_deprecated() noexcept:
    return 1

from scipy._lib.deprecation import deprecate_cython_api
import scipy._lib._test_deprecation_def as mod
deprecate_cython_api(mod, "foo_deprecated", new_name="foo",
                     message="Deprecated in Scipy 42.0.0")
del deprecate_cython_api, mod



# cython: cpow=True

# Implement Spence's function, a.k.a. the dilogarithm, for complex
# arguments. Note that our definition differs from that in the sources
# by the mapping z -> 1 - z.
#
# Sources
# [1] Zagier, "The Dilogarithm Function"
# [2] functions.wolfram.com
# [3] Ginsberg, Zaborowski, "The Dilogarithm Function of a Real Argument"
#
# Author: Josh Wilson
#
# Released under the same license as Scipy.

import cython
from ._complexstuff cimport zlog1, zabs

# Relative tolerance for the series
DEF TOL = 2.220446092504131e-16
DEF PISQ_6 = 1.6449340668482264365


@cython.cdivision(True)
cdef inline double complex cspence(double complex z) noexcept nogil:
    """
    Compute Spence's function for complex arguments. The strategy is:
    - If z is close to 0, use a series centered at 0.
    - If z is far away from 1, use the reflection formula

    spence(z) = -spence(z/(z - 1)) - pi**2/6 - ln(z - 1)**2/2

    to move close to 1. See [1].
    - If z is close to 1, use a series centered at 1.

    """
    if zabs(z) < 0.5:
        # This step isn't necessary, but this series converges faster.
        return cspence_series0(z)
    elif zabs(1 - z) > 1:
        return -cspence_series1(z/(z - 1)) - PISQ_6 - 0.5*zlog1(z - 1)**2
    else:
        return cspence_series1(z)


@cython.cdivision(True)
cdef inline double complex cspence_series0(double complex z) noexcept nogil:
    """
    A series centered at z = 0; see

    http://functions.wolfram.com/10.07.06.0005.02

    """
    cdef:
        int n
        double complex zfac = 1
        double complex sum1 = 0
        double complex sum2 = 0
        double complex term1, term2

    if z == 0:
        return PISQ_6

    for n in range(1, 500):
        zfac *= z
        term1 = zfac/n**2
        sum1 += term1
        term2 = zfac/n
        sum2 += term2
        if zabs(term1) <= TOL*zabs(sum1) and zabs(term2) <= TOL*zabs(sum2):
            break
    return PISQ_6 - sum1 + zlog1(z)*sum2


@cython.cdivision(True)
cdef inline double complex cspence_series1(double complex z) noexcept nogil:
    """
    A series centered at z = 1 which enjoys faster convergence than
    the Taylor series. See [3]. The number of terms used comes from
    bounding the absolute tolerance at the edge of the radius of
    convergence where the sum is O(1).

    """
    cdef:
        int n
        double complex zfac = 1
        double complex res = 0
        double complex term, zz

    if z == 1:
        return 0
    z = 1 - z
    zz = z**2
    for n in range(1, 500):
        zfac *= z
        # Do the divisions one at a time to guard against overflow
        term = ((zfac/n**2)/(n + 1)**2)/(n + 2)**2
        res += term
        if zabs(term) <= TOL*zabs(res):
            break
    res *= 4*zz
    res += 4*z + 5.75*zz + 3*(1 - zz)*zlog1(1 - z)
    res /= 1 + 4*z + zz
    return res



from importlib import import_module

import pytest

wrapped_libraries = ["numpy", "cupy", "torch", "dask.array"]
all_libraries = wrapped_libraries + [
    "array_api_strict", "jax.numpy", "ndonnx", "sparse"
]

def import_(library, wrapper=False):
    pytest.importorskip(library)
    if wrapper:
        if 'jax' in library:
            # JAX v0.4.32 implements the array API directly in jax.numpy
            # Older jax versions use jax.experimental.array_api
            jax_numpy = import_module("jax.numpy")
            if not hasattr(jax_numpy, "__array_api_version__"):
                library = 'jax.experimental.array_api'
        elif library in wrapped_libraries:
            library = 'array_api_compat.' + library

    return import_module(library)


def xfail(request: pytest.FixtureRequest, reason: str) -> None:
    """
    XFAIL the currently running test.

    Unlike ``pytest.xfail``, allow rest of test to execute instead of immediately
    halting it, so that it may result in a XPASS.
    xref https://github.com/pandas-dev/pandas/issues/38902
    """
    request.node.add_marker(pytest.mark.xfail(reason=reason))



from numpy import atleast_2d, arange, sum, cos, exp, pi
from .go_benchmark import Benchmark


class Watson(Benchmark):

    r"""
    Watson objective function.

    This class defines the Watson [1]_ global optimization problem. This is a
    unimodal minimization problem defined as follows:

    .. math::

        f_{\text{Watson}}(x) = \sum_{i=0}^{29} \left\{
                               \sum_{j=0}^4 ((j + 1)a_i^j x_{j+1})
                               - \left[ \sum_{j=0}^5 a_i^j
                               x_{j+1} \right ]^2 - 1 \right\}^2
                               + x_1^2


    Where, in this exercise, :math:`a_i = i/29`.

    with :math:`x_i \in [-5, 5]` for :math:`i = 1, ..., 6`.

    *Global optimum*: :math:`f(x) = 0.002288` for
    :math:`x = [-0.0158, 1.012, -0.2329, 1.260, -1.513, 0.9928]`

    .. [1] Jamil, M. & Yang, X.-S. A Literature Survey of Benchmark Functions
    For Global Optimization Problems Int. Journal of Mathematical Modelling
    and Numerical Optimisation, 2013, 4, 150-194.

    TODO Jamil #161 writes equation using (j - 1).  According to code in Adorio
    and Gavana it should be (j+1). However the equations in those papers
    contain (j - 1) as well.  However, I've got the right global minimum!!!
    """

    def __init__(self, dimensions=6):
        Benchmark.__init__(self, dimensions)

        self._bounds = list(zip([-5.0] * self.N, [5.0] * self.N))

        self.global_optimum = [[-0.0158, 1.012, -0.2329, 1.260, -1.513,
                                0.9928]]
        self.fglob = 0.002288

    def fun(self, x, *args):
        self.nfev += 1

        i = atleast_2d(arange(30.)).T
        a = i / 29.
        j = arange(5.)
        k = arange(6.)

        t1 = sum((j + 1) * a ** j * x[1:], axis=1)
        t2 = sum(a ** k * x, axis=1)

        inner = (t1 - t2 ** 2 - 1) ** 2

        return sum(inner) + x[0] ** 2


class Wavy(Benchmark):

    r"""
    Wavy objective function.

    This class defines the W / Wavy [1]_ global optimization problem. This is a
    multimodal minimization problem defined as follows:

    .. math::

        f_{\text{Wavy}}(x) = 1 - \frac{1}{n} \sum_{i=1}^{n}
                             \cos(kx_i)e^{-\frac{x_i^2}{2}}


    Where, in this exercise, :math:`k = 10`. The number of local minima is
    :math:`kn` and :math:`(k + 1)n` for odd and even :math:`k` respectively.

    Here, :math:`x_i \in [-\pi, \pi]` for :math:`i = 1, 2`.

    *Global optimum*: :math:`f(x) = 0` for :math:`x = [0, 0]`

    .. [1] Jamil, M. & Yang, X.-S. A Literature Survey of Benchmark Functions
    For Global Optimization Problems Int. Journal of Mathematical Modelling
    and Numerical Optimisation, 2013, 4, 150-194.
    """
    change_dimensionality = True

    def __init__(self, dimensions=2):
        Benchmark.__init__(self, dimensions)

        self._bounds = list(zip([-pi] * self.N, [pi] * self.N))

        self.global_optimum = [[0.0 for _ in range(self.N)]]
        self.fglob = 0.0

    def fun(self, x, *args):
        self.nfev += 1

        return 1.0 - (1.0 / self.N) * sum(cos(10 * x) * exp(-x ** 2.0 / 2.0))


class WayburnSeader01(Benchmark):

    r"""
    Wayburn and Seader 1 objective function.

    This class defines the Wayburn and Seader 1 [1]_ global optimization
    problem. This is a unimodal minimization problem defined as follows:

    .. math::

        f_{\text{WayburnSeader01}}(x) = (x_1^6 + x_2^4 - 17)^2
                                        + (2x_1 + x_2 - 4)^2


    with :math:`x_i \in [-5, 5]` for :math:`i = 1, 2`.

    *Global optimum*: :math:`f(x) = 0` for :math:`x = [1, 2]`

    .. [1] Jamil, M. & Yang, X.-S. A Literature Survey of Benchmark Functions
    For Global Optimization Problems Int. Journal of Mathematical Modelling
    and Numerical Optimisation, 2013, 4, 150-194.
    """

    def __init__(self, dimensions=2):
        Benchmark.__init__(self, dimensions)

        self._bounds = list(zip([-5.0] * self.N, [5.0] * self.N))
        self.custom_bounds = ([-2, 2], [-2, 2])

        self.global_optimum = [[1.0, 2.0]]
        self.fglob = 0.0

    def fun(self, x, *args):
        self.nfev += 1

        return (x[0] ** 6 + x[1] ** 4 - 17) ** 2 + (2 * x[0] + x[1] - 4) ** 2


class WayburnSeader02(Benchmark):

    r"""
    Wayburn and Seader 2 objective function.

    This class defines the Wayburn and Seader 2 [1]_ global optimization
    problem. This is a unimodal minimization problem defined as follows:

    .. math::

        f_{\text{WayburnSeader02}}(x) = \left[ 1.613 - 4(x_1 - 0.3125)^2
                                        - 4(x_2 - 1.625)^2 \right]^2
                                        + (x_2 - 1)^2


    with :math:`x_i \in [-500, 500]` for :math:`i = 1, 2`.

    *Global optimum*: :math:`f(x) = 0` for :math:`x = [0.2, 1]`

    .. [1] Jamil, M. & Yang, X.-S. A Literature Survey of Benchmark Functions
    For Global Optimization Problems Int. Journal of Mathematical Modelling
    and Numerical Optimisation, 2013, 4, 150-194.
    """

    def __init__(self, dimensions=2):
        Benchmark.__init__(self, dimensions)

        self._bounds = list(zip([-500.0] * self.N,
                           [500.0] * self.N))
        self.custom_bounds = ([-1, 2], [-1, 2])

        self.global_optimum = [[0.2, 1.0]]
        self.fglob = 0.0

    def fun(self, x, *args):
        self.nfev += 1

        u = (1.613 - 4 * (x[0] - 0.3125) ** 2 - 4 * (x[1] - 1.625) ** 2) ** 2
        v = (x[1] - 1) ** 2
        return u + v


class Weierstrass(Benchmark):

    r"""
    Weierstrass objective function.

    This class defines the Weierstrass [1]_ global optimization problem.
    This is a multimodal minimization problem defined as follows:

    .. math::

       f_{\text{Weierstrass}}(x) = \sum_{i=1}^{n} \left [
                                   \sum_{k=0}^{kmax} a^k \cos 
                                   \left( 2 \pi b^k (x_i + 0.5) \right) - n
                                   \sum_{k=0}^{kmax} a^k \cos(\pi b^k) \right ]


    Where, in this exercise, :math:`kmax = 20`, :math:`a = 0.5` and
    :math:`b = 3`.

    Here, :math:`n` represents the number of dimensions and
    :math:`x_i \in [-0.5, 0.5]` for :math:`i = 1, ..., n`.

    *Global optimum*: :math:`f(x) = 4` for :math:`x_i = 0` for
    :math:`i = 1, ..., n`

    .. [1] Mishra, S. Global Optimization by Differential Evolution and
    Particle Swarm Methods: Evaluation on Some Benchmark Functions.
    Munich Personal RePEc Archive, 2006, 1005

    TODO line 1591.
    TODO Jamil, Gavana have got it wrong.  The second term is not supposed to
    be included in the outer sum. Mishra code has it right as does the
    reference referred to in Jamil#166.
    """
    change_dimensionality = True

    def __init__(self, dimensions=2):
        Benchmark.__init__(self, dimensions)

        self._bounds = list(zip([-0.5] * self.N, [0.5] * self.N))

        self.global_optimum = [[0.0 for _ in range(self.N)]]
        self.fglob = 0

    def fun(self, x, *args):
        self.nfev += 1

        kmax = 20
        a, b = 0.5, 3.0

        k = atleast_2d(arange(kmax + 1.)).T
        t1 = a ** k * cos(2 * pi * b ** k * (x + 0.5))
        t2 = self.N * sum(a ** k.T * cos(pi * b ** k.T))

        return sum(sum(t1, axis=0)) - t2


class Whitley(Benchmark):

    r"""
    Whitley objective function.

    This class defines the Whitley [1]_ global optimization problem. This
    is a multimodal minimization problem defined as follows:

    .. math::

        f_{\text{Whitley}}(x) = \sum_{i=1}^n \sum_{j=1}^n
                                \left[\frac{(100(x_i^2-x_j)^2
                                + (1-x_j)^2)^2}{4000} - \cos(100(x_i^2-x_j)^2
                                + (1-x_j)^2)+1 \right]


    Here, :math:`n` represents the number of dimensions and
    :math:`x_i \in [-10.24, 10.24]` for :math:`i = 1, ..., n`.

    *Global optimum*: :math:`f(x) = 0` for :math:`x_i = 1` for
    :math:`i = 1, ..., n`

    .. [1] Gavana, A. Global Optimization Benchmarks and AMPGO retrieved 2015

    TODO Jamil#167 has '+ 1' inside the cos term, when it should be outside it.
    """
    change_dimensionality = True

    def __init__(self, dimensions=2):
        Benchmark.__init__(self, dimensions)

        self._bounds = list(zip([-10.24] * self.N,
                           [10.24] * self.N))
        self.custom_bounds = ([-1, 2], [-1, 2])

        self.global_optimum = [[1.0 for _ in range(self.N)]]
        self.fglob = 0.0

    def fun(self, x, *args):
        self.nfev += 1

        XI = x
        XJ = atleast_2d(x).T

        temp = 100.0 * ((XI ** 2.0) - XJ) + (1.0 - XJ) ** 2.0
        inner = (temp ** 2.0 / 4000.0) - cos(temp) + 1.0
        return sum(sum(inner, axis=0))


class Wolfe(Benchmark):

    r"""
    Wolfe objective function.

    This class defines the Wolfe [1]_ global optimization problem. This
    is a multimodal minimization problem defined as follows:

    .. math::

       f_{\text{Wolfe}}(x) = \frac{4}{3}(x_1^2 + x_2^2 - x_1x_2)^{0.75} + x_3


    with :math:`x_i \in [0, 2]` for :math:`i = 1, 2, 3`.

    *Global optimum*: :math:`f(x) = 0` for :math:`x = [0, 0, 0]`

    .. [1] Jamil, M. & Yang, X.-S. A Literature Survey of Benchmark Functions
    For Global Optimization Problems Int. Journal of Mathematical Modelling
    and Numerical Optimisation, 2013, 4, 150-194.
    """

    def __init__(self, dimensions=3):
        Benchmark.__init__(self, dimensions)

        self._bounds = list(zip([0.0] * self.N, [2.0] * self.N))

        self.global_optimum = [[0.0 for _ in range(self.N)]]
        self.fglob = 0.0

    def fun(self, x, *args):
        self.nfev += 1

        return 4 / 3 * (x[0] ** 2 + x[1] ** 2 - x[0] * x[1]) ** 0.75 + x[2]



import numpy as np
import matplotlib.pyplot as plt

from scipy import ndimage

img = np.array([-2, -1, 0, 1, 2], float)
x = np.linspace(-2, 6, num=1000)

modes = ['constant', 'grid-constant', 'nearest', 'reflect', 'mirror', 'wrap',
         'grid-wrap']
fig, axes = plt.subplots(len(modes), 3, figsize=(11, 8), sharex=True,
                         sharey=True)


for mode, (ax0, ax1, ax2) in zip(modes, axes):

    y = ndimage.map_coordinates(img, [x], order=0, mode=mode)
    ax0.scatter(np.arange(img.size), img)
    ax0.plot(x, y, '-')
    ax0.set_title(f'mode={mode}, order=0')

    y2 = ndimage.map_coordinates(img, [x], order=1, mode=mode)
    ax1.scatter(np.arange(img.size), img)
    ax1.plot(x, y2, '-')
    ax1.set_title(f'mode={mode}, order=1')

    y3 = ndimage.map_coordinates(img, [x], order=3, mode=mode)
    ax2.scatter(np.arange(img.size), img)
    ax2.plot(x, y3, '-')
    ax2.set_title(f'mode={mode}, order=3')

    sz = len(img)
    for ax in (ax0, ax1, ax2):
        if mode in ['grid-wrap', 'reflect']:
            ax.plot([-0.5, -0.5], [-2.5, 2.5], 'k--')
            ax.plot([sz - 0.5, sz - 0.5], [-2.5, 2.5], 'k--')
        elif mode in ['wrap', 'mirror']:
            ax.plot([0, 0], [-2.5, 2.5], 'k--')
            ax.plot([sz - 1, sz - 1], [-2.5, 2.5], 'k--')

    if mode != 'constant':
        for xx in range(int(x[0]), int(x[-1] + 1)):
            if (xx < 0) or (xx > img.size - 1):
                idx = np.argmin(np.abs(x - xx))

                for y_vals, ax in zip((y, y2, y3), (ax0, ax1, ax2)):
                    ax.scatter(
                        [x[idx]], [y_vals[idx]], facecolors='none',
                        edgecolor='#0343df', marker='o'
                    )

plt.tight_layout()
plt.show()



import matplotlib.pyplot as plt
import numpy as np

aa = [1, 1, 2, 2]  # amplitudes
taus = [1, 2, 1, 2]  # durations

fg0, axx = plt.subplots(3, 1, sharex='all', tight_layout=True, figsize=(6., 4.))
axx[0].set(title=r"Spectrum $|X(f)|$", ylabel="V/Hz")
axx[1].set(title=r"Magnitude Spectrum $|X(f)/\tau|$ ", ylabel=r"V")
axx[2].set(title=r"Amplitude Spectral Density $|X(f)/\sqrt{\tau}|$",
           ylabel=r"$\operatorname{V} / \sqrt{\operatorname{Hz}}$",
           xlabel="Frequency $f$ in Hertz",)

x_labels, x_ticks = [], []
f = np.linspace(-2.5, 2.5, 400)
for c_, (a_, tau_) in enumerate(zip(aa, taus), start=1):
    aZ_, f_ = abs(a_ * tau_ * np.sinc(tau_ * f) / 2), f + c_ * 5
    axx[0].plot(f_, aZ_)
    axx[1].plot(f_, aZ_ / tau_)
    axx[2].plot(f_, aZ_ / np.sqrt(tau_))
    x_labels.append(rf"$a={a_:g}$, $\tau={tau_:g}$")
    x_ticks.append(c_ * 5)

axx[2].set_xticks(x_ticks)
axx[2].set_xticklabels(x_labels)
plt.show()



# cython: language_level=3
# cython: cdivision=True
import importlib.resources

cimport cython
cimport numpy as cnp

import numpy as np

cnp.import_array()

# Parameters are linked to the direction numbers list.
# See `_initialize_direction_numbers` for more details.
# Declared using DEF to be known at compilation time for ``poly`` et ``vinit``
DEF MAXDIM = 21201  # max number of dimensions
DEF MAXDEG = 18  # max polynomial degree


ctypedef fused uint_32_64:
    cnp.uint32_t
    cnp.uint64_t


# Needed to be accessed with python
cdef extern from *:
    """
    int MAXDIM_DEFINE = 21201;
    int MAXDEG_DEFINE = 18;
    """
    int MAXDIM_DEFINE  # max number of dimensions
    int MAXDEG_DEFINE  # max polynomial degree

_MAXDIM = MAXDIM_DEFINE
_MAXDEG = MAXDEG_DEFINE

_poly_dict = {}
_vinit_dict = {}


def get_poly_vinit(kind, dtype):
    """Initialize and cache the direction numbers.

    Uses a dictionary to store the arrays. `kind` allows to select which
    dictionary to pull. The key of each dictionary corresponds to the `dtype`.
    If the key is not present in any of the dictionary, both dictionaries are
    initialized with `_initialize_direction_numbers`, for the given `dtype`.

    This is only used during the initialization step in `_initialize_v`.

    Parameters
    ----------
    kind : {'poly', 'vinit'}
        Select which dictionary to pull.
    dtype : {np.uint32, np.uint64}
        Which dtype to use.

    Returns
    -------
    poly_vinit : np.ndarray
        Either ``poly`` or ``vinit`` matrix.

    """
    if kind == 'poly':
        poly_vinit = _poly_dict.get(dtype)
    else:
        poly_vinit = _vinit_dict.get(dtype)

    if poly_vinit is None:
        poly = np.empty((MAXDIM,), dtype=dtype)
        vinit = np.empty((MAXDIM, MAXDEG), dtype=dtype)

        _initialize_direction_numbers(poly, vinit, dtype)

        _poly_dict[dtype] = poly
        _vinit_dict[dtype] = vinit

        if kind == 'poly':
            poly_vinit = _poly_dict.get(dtype)
        else:
            poly_vinit = _vinit_dict.get(dtype)

    return poly_vinit


def _initialize_direction_numbers(poly, vinit, dtype):
    """Load direction numbers into two arrays.

    Parameters
    ----------
    poly, vinit : np.ndarray
        Direction numbers arrays to fill.
    dtype : {np.uint32, np.uint64}
        Which dtype to use.

    Notes
    -----
    Direction numbers obtained using the search criterion D(6)
    up to the dimension 21201. This is the recommended choice by the authors.

    Original data can be found at https://web.maths.unsw.edu.au/~fkuo/sobol/.
    For additional details on the quantities involved, see [1].

    [1] S. Joe and F. Y. Kuo. Remark on algorithm 659: Implementing sobol's
        quasirandom sequence generator. ACM Trans. Math. Softw., 29(1):49-57,
        Mar. 2003.

    The C-code generated from putting the numbers in as literals is obscenely
    large/inefficient. The data file was thus packaged and save as an .npz data
    file for fast loading using the following code (this assumes that the file
    https://web.maths.unsw.edu.au/~fkuo/sobol/new-joe-kuo-6.21201 is present in
    the working directory):

        import pandas as pd
        import numpy as np

        # read in file content
        with open("./new-joe-kuo-6.21201", "r") as f:
            lines = f.readlines()

        rows = []

        # parse data from file line by line
        for l in lines[1:]:
            nums = [int(n) for n in l.replace(" \n", "").split()]
            d, s, a = nums[:3]
            vs = {f"v{i}": int(v) for i,v in enumerate(nums[3:])}
            rows.append({"d": d, "s": s, "a": a, **vs})


        # read in as dataframe, explicitly use zero values
        df = pd.DataFrame(rows).fillna(0).astype(int)

        # perform conversion
        df["poly"] = 2 * df["a"] + 2 ** df["s"] + 1

        # ensure columns are properly ordered
        vs = df[[f"v{i}" for i in range(18)]].values

        # add the degenerate d=1 column (not included in the data file)
        vs = np.vstack([vs[0][np.newaxis, :], vs])
        poly = np.concatenate([[1], df["poly"].values])

        # save as compressed .npz file to minimize size of distribution
        np.savez_compressed("./_sobol_direction_numbers", vinit=vs, poly=poly)

    """
    _curdir = importlib.resources.files("scipy.stats")
    _npzfile = _curdir.joinpath("_sobol_direction_numbers.npz")
    with importlib.resources.as_file(_npzfile) as f:
        dns = np.load(f)

    dns_poly = dns["poly"].astype(dtype)
    dns_vinit = dns["vinit"].astype(dtype)
    poly[...] = dns_poly
    vinit[...] = dns_vinit


@cython.boundscheck(False)
@cython.wraparound(False)
cdef int bit_length(uint_32_64 n) noexcept:
    cdef int bits = 0
    cdef uint_32_64 nloc = n
    while nloc != 0:
        nloc >>= 1
        bits += 1
    return bits


@cython.boundscheck(False)
@cython.wraparound(False)
cdef int low_0_bit(uint_32_64 x) noexcept nogil:
    """Get the position of the right-most 0 bit for an integer.

    Examples:
    >>> low_0_bit(0)
    1
    >>> low_0_bit(1)
    2
    >>> low_0_bit(2)
    1
    >>> low_0_bit(5)
    2
    >>> low_0_bit(7)
    4

    Parameters
    ----------
    x : int
        An integer.

    Returns
    -------
    position : int
        Position of the right-most 0 bit.

    """
    cdef int i = 0
    while x & (1 << i) != 0:
        i += 1
    return i + 1


@cython.boundscheck(False)
@cython.wraparound(False)
cdef int ibits(uint_32_64 x, const int pos, const int length) noexcept nogil:
    """Extract a sequence of bits from the bit representation of an integer.

    Extract the sequence from position `pos` (inclusive) to ``pos + length``
    (not inclusive), leftwise.

    Examples:
    >>> ibits(1, 0, 1)
    1
    >>> ibits(1, 1, 1)
    0
    >>> ibits(2, 0, 1)
    0
    >>> ibits(2, 0, 2)
    2
    >>> ibits(25, 1, 5)
    12


    Parameters
    ----------
    x : int
        Integer to convert to bit representation.
    pos : int
        Starting position of sequence in bit representation of integer.
    length : int
        Length of sequence (number of bits).

    Returns
    -------
    ibits : int
        Integer value corresponding to bit sequence.

    """
    return (x >> pos) & ((1 << length) - 1)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void _initialize_v(
    uint_32_64[:, ::1] v, const int dim, const int bits
) noexcept:
    """Initialize matrix of size ``dim * bits`` with direction numbers."""
    cdef int d, i, j, k, m
    cdef uint_32_64 p, newv, pow2
    cdef uint_32_64[:] poly = get_poly_vinit(
        'poly',
        np.uint32 if uint_32_64 is cnp.uint32_t else np.uint64
    )
    cdef uint_32_64[:, ::1] vinit = get_poly_vinit(
        'vinit',
        np.uint32 if uint_32_64 is cnp.uint32_t else np.uint64
    )

    if dim == 0:
        return

    # first row of v is all 1s
    for i in range(bits):
        v[0, i] = 1

    # Remaining rows of v (row 2 through dim, indexed by [1:dim])
    for d in range(1, dim):
        p = poly[d]
        m = bit_length(p) - 1

        # First m elements of row d comes from vinit
        v[d, :m] = vinit[d, :m]

        # Fill in remaining elements of v as in Section 2 (top of pg. 90) of:
        #
        # P. Bratley and B. L. Fox. Algorithm 659: Implementing sobol's
        # quasirandom sequence generator. ACM Trans.
        # Math. Softw., 14(1):88-100, Mar. 1988.
        #
        for j in range(m, bits):
            newv = v[d, j - m]
            pow2 = 1
            for k in range(m):
                pow2 = pow2 << 1
                if (p >> (m - 1 - k)) & 1:
                    newv = newv ^ (pow2 * v[d, j - k - 1])
            v[d, j] = newv

    # Multiply each column of v by power of 2:
    # v * [2^(bits-1), 2^(bits-2),..., 2, 1]
    pow2 = 1
    for d in range(bits):
        for i in range(dim):
            v[i, bits - 1 - d] *= pow2
        pow2 = pow2 << 1


def _draw(
    n,
    num_gen,
    const int dim,
    const cnp.float64_t scale,
    const uint_32_64[:, ::1] sv,
    uint_32_64[::1] quasi,
    cnp.float64_t[:, ::1] sample
):
    # necessary wrapper to guide Cython for n, num_gen and scale
    cdef uint_32_64 n_ = n
    cdef uint_32_64 num_gen_ = num_gen
    draw(n_, num_gen_, dim, scale, sv, quasi, sample)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void draw(
    const uint_32_64 n,
    const uint_32_64 num_gen,
    const int dim,
    const cnp.float64_t scale,
    const uint_32_64[:, ::1] sv,
    uint_32_64[::1] quasi,
    cnp.float64_t[:, ::1] sample
) noexcept nogil:
    cdef int j, l
    cdef uint_32_64 num_gen_loc = num_gen
    cdef uint_32_64 i, qtmp

    for i in range(n):
        l = low_0_bit(num_gen_loc)
        for j in range(dim):
            qtmp = quasi[j] ^ sv[j, l - 1]
            quasi[j] = qtmp
            sample[i, j] = qtmp * scale
        num_gen_loc += 1


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void _fast_forward(const uint_32_64 n,
                         const uint_32_64 num_gen,
                         const int dim,
                         const uint_32_64[:, ::1] sv,
                         uint_32_64[::1] quasi) noexcept nogil:
    cdef int j, l
    cdef uint_32_64 num_gen_loc = num_gen
    cdef uint_32_64 i
    for i in range(n):
        l = low_0_bit(num_gen_loc)
        for j in range(dim):
            quasi[j] = quasi[j] ^ sv[j, l - 1]
        num_gen_loc += 1


@cython.boundscheck(False)
@cython.wraparound(False)
cdef uint_32_64 cdot_pow2(const uint_32_64[::1] a) noexcept nogil:
    cdef int i
    cdef int size = a.shape[0]
    cdef uint_32_64 z = 0
    cdef uint_32_64 pow2 = 1
    for i in range(size):
        z += a[size - 1 - i] * pow2
        pow2 *= 2
    return z


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void _cscramble(const int dim,
                      const int bits,
                      uint_32_64[:, :, ::1] ltm,
                      uint_32_64[:, ::1] sv) noexcept nogil:
    """Scrambling using (left) linear matrix scramble (LMS)."""
    cdef int d, i, j, k, p
    cdef uint_32_64 l, lsmdp, t1, t2, vdj

    # Set diagonals of bits x bits arrays to 1
    for d in range(dim):
        for i in range(bits):
            ltm[d, i, i] = 1

    for d in range(dim):
        for j in range(bits):
            vdj = sv[d, j]
            l = 1
            t2 = 0
            for p in range(bits - 1, -1, -1):
                lsmdp = cdot_pow2(ltm[d, p, :])
                t1 = 0
                for k in range(bits):
                    t1 += ibits(lsmdp, k, 1) * ibits(vdj, k, 1)
                t1 = t1 % 2
                t2 = t2 + t1 * l
                l = 2 * l
            sv[d, j] = t2


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void _fill_p_cumulative(const cnp.float_t[::1] p,
                              cnp.float_t[::1] p_cumulative) noexcept nogil:
    cdef int i
    cdef int len_p = p.shape[0]
    cdef float tot = 0
    cdef float t
    for i in range(len_p):
        t = tot + p[i]
        p_cumulative[i] = t
        tot = t


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void _categorize(const cnp.float_t[::1] draws,
                       const cnp.float_t[::1] p_cumulative,
                       cnp.intp_t[::1] result) noexcept nogil:
    cdef int i
    cdef int n_p = p_cumulative.shape[0]
    for i in range(draws.shape[0]):
        j = _find_index(p_cumulative, n_p, draws[i])
        result[j] = result[j] + 1


@cython.boundscheck(False)
@cython.wraparound(False)
cdef int _find_index(const cnp.float_t[::1] p_cumulative,
                     const int size,
                     const float value) noexcept nogil:
    cdef int l = 0
    cdef int r = size - 1
    cdef int m
    while r > l:
        m = (l + r) // 2
        if value > p_cumulative[m]:
            l = m + 1
        else:
            r = m
    return r


def _test_find_index(p_cumulative, size, value):
    # type: (np.ndarray, int, float) -> int
    """Wrapper for testing in python"""
    return _find_index(p_cumulative, size, value)
